@@ -50,5 +50,5 @@ class Worker_Plot(QtCore.QThread):
         self.Project = project
 
     def run(self):
-        self.Project.plot_sankey(view=False, filename=os.getcwd()+'\\sankey.html')
+        self.Project.plot_sankey(view=False, filename=os.getcwd()+'\\sankey')
         self.Plot.emit()
