@@ -5,6 +5,7 @@ Created on Wed Jul 29 17:03:27 2020
 @author: msmsa
 """
 import numpy as np
+import pandas as pd
 from PFAS_SAT_ProcessModels.SubProcesses import split
 import graphviz
 import plotly.graph_objects as go
@@ -13,7 +14,7 @@ import json
 import warnings
 import time
 from .utils import NpEncoder
-import importlib  # to import moduls with string name
+import importlib  # to import module with string name
 from PFAS_SAT_ProcessModels import ProcessModelsMetaData
 
 
@@ -160,8 +161,12 @@ class Project():
         self.FlowParams = FlowParams
 
     def setup_network(self, Cut_Off=0.001):
+        self._CuttOff_factor = Cut_Off
         self.Inventory.clear()
-        self.CuttOff = sum(self.InputFlow.PFAS.values) * Cut_Off
+        if self.InputFlow.PFAS.values.sum() > 0:
+            self.CuttOff = self.InputFlow.PFAS.values.sum() * self._CuttOff_factor
+        else:
+            self.CuttOff = 0.001
         product = {self.InputFlow.FlowType: self.InputFlow}
         Project.calc(product, self.InputFlow.FlowType, self.ProcessModels, self.FlowParams, self.Inventory, self.CuttOff,
                      self._ProcessNameRef, self.WasteTreatment, pop_up=self.pop_up)
@@ -192,29 +197,31 @@ class Project():
                         pop_up('Treatment Network Warning!', msg, 'Warning')
                     warnings.warn(msg)
 
-    def setup_MC(self, InputFlow_object, seed=None):
+    def setup_MC(self, InputFlow_object, seed=None, flow_concentration=True):
         self.InputFlow_object = InputFlow_object
-        self.InputFlow_object.setup_MC(seed=seed)
+        self.InputFlow_object.setup_MC(seed=seed,
+                                       concentration=flow_concentration)
         for p in self.ProcessSet:
             self.ProcessModels[p]['Model'].setup_MC(seed=seed)
 
-    def MC_Next(self, InputFlow):
+    def MC_Next(self):
         NewInputData = []
-        if InputFlow:
-            NewInputData += self.InputFlow_object.MC_Next()
+        NewInputData += self.InputFlow_object.MC_Next()
         for p in self.ProcessSet:
             Raw_newdata = []
             for x in self.ProcessModels[p]['Model'].MC_Next():
                 Raw_newdata.append(((p, x[0]), x[1]))
             NewInputData += Raw_newdata
-
         self.Inventory.clear()
         self.InputFlow_object.calc()
+        self.CuttOff = self.InputFlow_object.Inc_flow.PFAS.values.sum() * self._CuttOff_factor
+        self.CuttOff = 0.001 if self.CuttOff == 0 else self.CuttOff
         product = {self.InputFlow_object.Inc_flow.FlowType: self.InputFlow_object.Inc_flow}
         Project.calc(product, 'Start', self.ProcessModels, self.FlowParams, self.Inventory, self.CuttOff, self._ProcessNameRef, self.WasteTreatment)
         return(NewInputData)
 
-    def MC_Run(self, n, InputFlow=True, TypeOfPFAS='All', signal=None):
+    def MC_Run(self, n, TypeOfPFAS='All', normalize=False,
+               Start_flow=None, signal=None):
         # Print warning only first time
         warnings.simplefilter('once', UserWarning)
 
@@ -223,8 +230,12 @@ class Project():
         progress_steps = [int(x) for x in np.linspace(1, n, 19)]
         MC_results = []
         for i in range(n):
-            NewInputData = self.MC_Next(InputFlow)
-            MC_results.append((i, NewInputData, self.Inventory.report(TypeOfPFAS)))
+            NewInputData = self.MC_Next()
+            MC_results.append((i,
+                               NewInputData,
+                               self.Inventory.report(TypeOfPFAS=TypeOfPFAS,
+                                                     normalize=normalize,
+                                                     Start_flow=Start_flow)))
             if i in progress_steps:
                 if signal:
                     progress += 5
@@ -245,14 +256,14 @@ class Project():
     def setup_SA(self, InputFlow_object):
         self.InputFlow_object = InputFlow_object
 
-    def SensitivityAnalysis(self, Model, Category, Paramter, Start=None,
-                            Stop=None, Nstep=20, TypeOfPFAS='All'):
+    def SensitivityAnalysis(self, Model, Category, Parameter, Start=None,
+                            Stop=None, Nstep=20, TypeOfPFAS='All', Default_Res=False):
         try:
             SA_results = []
             if Model == 'IncomFlow':
-                param_dict = self.InputFlow_object.InputData.Input_dict[Category][Paramter]
+                param_dict = self.InputFlow_object.InputData.Input_dict[Category][Parameter]
             else:
-                param_dict = self.ProcessModels[Model]['Model'].InputData.Input_dict[Category][Paramter]
+                param_dict = self.ProcessModels[Model]['Model'].InputData.Input_dict[Category][Parameter]
 
             Start = param_dict['minimum'] if Start is None else Start
             Stop = param_dict['maximum'] if Stop is None else Stop
@@ -260,18 +271,22 @@ class Project():
                 raise ValueError('Set the Start & Stop.')
 
             param_vals = np.linspace(Start, Stop, Nstep)
+            if Default_Res:
+                param_vals = [param_dict['amount']]
             for i, val in enumerate(param_vals):
                 # Update the Input data
                 param_dict['amount'] = val
                 # Calculating results
                 self.Inventory.clear()
                 self.InputFlow_object.calc()
+                self.CuttOff = self.InputFlow_object.Inc_flow.PFAS.values.sum() * self._CuttOff_factor
+                self.CuttOff = 0.001 if self.CuttOff == 0 else self.CuttOff
                 product = {self.InputFlow_object.Inc_flow.FlowType: self.InputFlow_object.Inc_flow}
                 Project.calc(product, 'Start', self.ProcessModels, self.FlowParams, self.Inventory, self.CuttOff,
                              self._ProcessNameRef, self.WasteTreatment)
 
                 # Store Results
-                SA_results.append((i, [((Category, Paramter), val)], self.Inventory.report(TypeOfPFAS)))
+                SA_results.append((i, [((Category, Parameter), val)], self.Inventory.report(TypeOfPFAS)))
 
         except Exception as e:
             print('Error while running the SA: \n')
@@ -281,7 +296,54 @@ class Project():
             # Reset Values
             self.Reset_static_Data()
             self.Inventory.clear()
-            return(SA_results)
+            return SA_results
+
+    def SA_SR(self, mc_res_cols, TypeOfPFAS, target_result):
+        uncertain_params = mc_res_cols[len(self.Inventory.REPORT_INDEX) + 1:]
+        target = target_result.replace(' (10e-6g)', '')
+        results = pd.DataFrame(columns=[f'SR {target}'], dtype=object)
+        for param in uncertain_params:
+            kwargs = {}
+            kwargs['Nstep'] = 2
+            kwargs['TypeOfPFAS'] = TypeOfPFAS
+            if isinstance(param[1], tuple):
+                kwargs['Model'] = param[0]
+                kwargs['Category'] = param[1][0]
+                kwargs['Parameter'] = param[1][1]
+                param_dict = self.ProcessModels[kwargs['Model']]['Model'].InputData.Input_dict[kwargs['Category']][kwargs['Parameter']]
+            else:
+                kwargs['Model'] = 'IncomFlow'
+                kwargs['Category'] = param[0]
+                kwargs['Parameter'] = param[1]
+                param_dict = self.InputFlow_object.InputData.Input_dict[kwargs['Category']][kwargs['Parameter']]
+
+            if not np.isnan(param_dict['maximum']):
+                kwargs['Stop'] = param_dict['maximum']
+            elif param_dict['uncertainty_type'] == 3:
+                kwargs['Stop'] = param_dict['loc'] + 3 * param_dict['scale']
+            else:
+                next
+
+            if not np.isnan(param_dict['minimum']):
+                kwargs['Start'] = param_dict['minimum']
+            elif param_dict['uncertainty_type'] == 3:
+                kwargs['Start'] = param_dict['loc'] - 3 * param_dict['scale']
+            else:
+                next
+
+            sa_res = self.SensitivityAnalysis(Default_Res=True, **kwargs)
+            res = sa_res[0][2][target_result]
+
+            sa_res_1 = self.SensitivityAnalysis(**kwargs)
+            res_min = sa_res_1[0][2][target_result]
+            res_max = sa_res_1[1][2][target_result]
+            try:
+                SR = ((res_max - res_min) / res) / ((kwargs['Stop'] - kwargs['Start']) / param_dict['amount'])
+                results.loc[str(param), f'SR {target}'] = SR
+            except Exception:
+                next
+        results.sort_values(by=f'SR {target}', key=abs, inplace=True, ascending=False)
+        return results
 
     def plot_network(self, view=True, show_vals=True, file_name='Network'):
         """
@@ -317,10 +379,19 @@ class Project():
                   Make sure that the directory containing the dot executable is on your systems’ path.
                   """)
 
-    def plot_sankey(self, view=True, filename=None):
-
+    def plot_sankey(self, view=True, filename=None, TypeOfPFAS='All', normalize=False, Start_flow=None):
         Data = self.Inventory.Inv
 
+        if TypeOfPFAS == 'All':
+            pfas_index = self.CommonData.PFAS_Index
+        else:
+            pfas_index = [TypeOfPFAS]
+
+        if normalize and Start_flow:
+            total_pfas = sum(Start_flow.PFAS[pfas_index].values)
+        else:
+            total_pfas = 1
+
         label = []
         label_dict = {}
         source = []
@@ -330,26 +401,47 @@ class Project():
         color_link = []
 
         # List of colors: https://flaviocopes.com/rgb-color-codes/
-        edge_color_dict = {'FoodWaste': (202, 255, 112),  # darkolivegreen1  #CAFF70
+        edge_color_dict = {'ADLiquids': (220, 20, 60),  # crimson    #DC143C
+                           'ADSolids': (124, 252, 0),  # lawn green    #7CFC00
+                           'AFFF': (138, 43, 226),  # blue violet	#8A2BE2
+                           'C_DWaste': (238, 232, 205),  # cornsilk2    #EEE8CD
+                           'CombustionResiduals': (169, 169, 169),  # darkgray    #A9A9A9
                            'Compost': (110, 139, 61),  # darkolivegreen4    #6E8B3D
                            'CompostResiduals': (118, 238, 0),  # chartreuse2    #76EE00
-                           'ADSolids': (124, 252, 0),  # lawn green    #7CFC00
-                           'MSW': (139, 101, 8),  # darkgoldenrod4    #8B6508
-                           'C_DWaste': (238, 232, 205),  # cornsilk2    #EEE8CD
-                           'ADLiquids': (220, 20, 60),  # crimson    #DC143C
-                           'SpentGAC': (189, 183, 107),  # darkkhaki    #BDB76B
-                           'ContaminatedSoil': (128, 138, 135),  # coldgrey    #808A87
-                           'CombustionResiduals': (169, 169, 169),  # darkgray    #A9A9A9
+                           'Contact Water': (142, 229, 238),  # cadetblue2    #8EE5EE
                            'ContactWater': (142, 229, 238),  # cadetblue2    #8EE5EE
+                           'ContaminatedSoil': (128, 138, 135),  # coldgrey    #808A87
                            'ContaminatedWater': (0, 0, 139),  # dark blue    #00008B
-                           'LFLeachate': (0, 0, 255),  # blue    #0000FF
+                           'Deep Well Injection': (65, 105, 225),  # royal blue	#4169E1
+                           'Mineralized': (255, 165, 0),  # orange	#FFA500
+                           'Destructed': (255, 165, 0),  # orange	#FFA500
                            'DewateredWWTSolids': (128, 128, 128),  # gray / grey    #808080
                            'DriedWWTSolids': (192, 192, 192),  # silver    #C0C0C0
-                           'RawWWTSolids': (112, 128, 144),  # slate gray    #708090
+                           'DryerExhaust': (128, 128, 128),  # gray / grey	#808080
+                           'Effluent': (0, 191, 255),  # deep sky blue	#00BFFF
+                           'Exhaust': (128, 128, 128),  # gray / grey	#808080
+                           'FoodWaste': (202, 255, 112),  # darkolivegreen1  #CAFF70
+                           'GAC Effluent': (72, 209, 204),  # medium turquoise	#48D1CC
+                           'IER Effluent': (72, 209, 204),  # medium turquoise	#48D1CC
+                           'LFLeachate': (0, 0, 255),  # blue    #0000FF
+                           'Leachate': (106, 90, 205),  # slate blue	#6A5ACD
+                           'MSW': (139, 101, 8),  # darkgoldenrod4    #8B6508
+                           'RO Effluent': (72, 209, 204),  # medium turquoise	#48D1CC
                            'ROConcentrate': (123, 104, 238),  # medium slate blue    #7B68EE
+                           'RawWWTSolids': (112, 128, 144),  # slate gray    #708090
+                           'Reactivated GAC': (220, 20, 60),  # crimson	#DC143C
+                           'Release': (32, 178, 170),  # light sea green	#20B2AA
+                           'Remaining': (244, 164, 96),  # sandy brown	#F4A460
+                           'RunOff': (0, 255, 255),  # aqua    #00FFFF
+                           'SpentGAC': (189, 183, 107),  # darkkhaki    #BDB76B
+                           'SpentIER': (189, 183, 107),  # darkkhaki    #BDB76B
+                           'Stabilized': (218, 165, 32),  # golden rod	#DAA520
                            'StabilizedSoil': (72, 61, 139),  # dark slate blue    #483D8B
-                           'ROConc': (70, 130, 180),  # steel blue    #4682B4
-                           'RunOff': (0, 255, 255)}  # aqua    #00FFFF
+                           'Storage': (233, 150, 122),  # dark salmon	#E9967A
+                           'Volatilization': (255, 127, 80),  # coral	#FF7F50
+                           'Volatilized': (255, 127, 80),  # coral	#FF7F50
+                           'WWTEffluent': (102, 205, 170),  # medium aqua marine	#66CDAA
+                           'WWTScreenRejects': (255, 99, 71)}  # tomato	#FF6347
 
         for i in Data.loc['Flow_name']:
             if i not in edge_color_dict:
@@ -370,8 +462,8 @@ class Project():
                 index += 1
             target.append(label_dict[i])
 
-        for i in Data.loc[self.CommonData.PFAS_Index].sum():
-            value.append(i)
+        for i in Data.loc[pfas_index].sum():
+            value.append(i / total_pfas * 100)
 
         for i in Data.loc['Flow_name']:
             label_link.append(i)
@@ -407,10 +499,16 @@ class Project():
         layout = go.Layout(title_text=None,
                            font_size=16,
                            hoverlabel=dict(font_size=14))
-        data = go.Sankey(valueformat=".3s",
-                         valuesuffix="μg",
-                         node=node,
-                         link=link)
+        if normalize and Start_flow:
+            data = go.Sankey(valueformat=".4s",
+                             valuesuffix="%",
+                             node=node,
+                             link=link)
+        else:
+            data = go.Sankey(valueformat=".3s",
+                             valuesuffix="μg",
+                             node=node,
+                             link=link)
         fig = go.Figure(data=[data], layout=layout)
         plot(fig, filename=filename + '.html' if filename else 'sankey.html', auto_open=view)
 
